// vcs.e $Revision: 1.25 $
//
// 3 December 2002
//
// Purpose:
//
//		This extension is designed to interact with version control
//		systems.  A problem with most VCS systems is that in the
//		typical multi-programmer environment files are stored locally
//		as Read Only files.  This is only a problem in the sense that
//		the programmer must explicitly lock the file before editing
//		it.  This extension handles typical VCS interaction from the
//		editor including get, get and lock, put, differences with
//		current tip revisions, merge, merge and lock and current
//		revision history.
//
// Usage:
//
//		Compile and load this extension.  Add vcs.doc to the end if
//		EDOC and save the state.  See vcs.doc for a list of commands
//		and variables and their effect.  The only assumptions made by
//		this extensions are:
//
//			1) The environment variable for the particular system is
//				set (ie. SSUSER or TLIBID or VCSID or LOGNAME).
//			2) The VCS system is set up for Multi-Programmer lock
//				use with IDs and the local files are stored in read
//				only format when not checked out (a requirement for
//				most systems).
//			3) The VCS system is configured to automatically find
//				it's log files based on the directory in which the
//				file being edited resides.
//
//		If 1) is not true, correct it.  If 2) is not true, you do not
//		need this extension.  If 3) is not true, then you will need
//		to modify the cfg file for the system in use to make this true.
//
//		This extension is set up to allow the specification of a
//		default vcs project directory and it's associated project
//		name.  This is primarily to support the SourceSafe command
//		line which does not easily fit into assumption 3) above and
//		specifically assumes a hierarchical project arrangment.
//		Modification of VCSSetCurrentProject() is the place to change
//		this if required.
//
//		The vcs variables are stored in the session file to allow
//		project specific vcs settings.  This is especially nice for
//		those using more than one VCS system or for those in
//		transition from one to another.
//
// Details:
//
//		The basic thread of each command is to determine the current
//		VCS system, set the current project, save the current editing
//		environment, perform the command and restore the system.  In
//		order:
//
//			The current VCS systems is determined first by the value
//			of VCS_default.  If VCS_default is -1 then a search of
//			the environment is made and the first VCS system which
//			has it's ID in the environment will be used.  Search is
//			according to order in the vcsCommandList variable below.
//
//			The current project is set only when the vcsSetProject
//			member of the vcsCommandList is non zero in length.  The
//			variables VCS_project_base and VCS_project_name are used
//			to create the current project name by appending the path
//			of the file relative to the project base to the project
//			name.  If the file is not a descendent in the directory
//			tree from the vcs project base then an error occurs and
//			the user is informed.
//
//			The current point, mark, buffer and filename are stored.
//			If the file is modified, it is saved.  The current
//			directory is changed to the file location (to allow
//			configuration files in the local directory to be picked
//			up) and a temporary buffer is created to maintain the
//			window space of the buffer (for the current window only).
//			At the end of this, the current process buffer,
//			error_point and grep variables are saved.  The reason the
//			the process save is made is to facilitate changes
//			required as a result of compiler or grep searches without
//			interrupting the current search.  In addition, any
//			concurrent processes are suspended.
//
//			The commands are executed via a variant of do_push with
//			an appropriate value of shell_shrink to allow the maximum
//			amount of memory to the process.
//
//			Once the command is complete, the process buffers are
//			restored, the new file is re-read and placed in the saved
//			window and the concurrent shell if any are restored.  The
//			buffer point and mark are restored and the grep/compile
//			state is reset.
//
//		The command names used are primarily pneumonic in nature
//		(like get, put, branch and merge).  Those which were not
//		obvious were taken from the RCS/PVCS command set (since this
//		command set is by far the largest chunk of the market share),
//		thus vdiff (to avoid name collision with diff) and vlog.
//
//		The commands are not bound to any keys.  These are not really
//		keystroke operations and were forced to the A-X command line.
//		If you need them bound to keys, go right ahead.
//
// Some future directions:
//
//		1) Add more VCS systems.  In particular, perhaps Sourcerer's
//		Apprentice and Microsoft Delta.
//
// The modifications to this code are dedicated to the public domain
// with the caveat that Lugaru is welcome to use this within their
// distribution source code which is supplied with Epsilon.  Use it
// any way you want for whatever purposes.
//
//
//	John Kercheval
//  4002 NE 204th St
//  Lake Forest Park, WA  98155-1606
//	INTERNET: kercheval@bigfoot.com, johnk@wrq.com
//	Compu$erve: 72450,3702
//
//
// Modification History:
//
//		J. Kercheval  Tue, 01/21/1992 21:53:22   creation
//		J. Kercheval  Mon, 02/10/1992 14:49:09   add mergel
//		J. Kercheval  Thu, 09/24/1992 15:23:00   add process buffer save and restore
//		J. Kercheval  Fri, 06/18/1993 13:44:33   add -b and reverse file order to vdiff command
//		J. Kercheval  Fri, 06/18/1993 13:48:49   validate _VCSErrorSpot is in range
//		J. Kercheval  Wed, 07/14/1993 12:15:43   remove dependency on external non-PVCS utilities
//		J. Kercheval  Thu, 07/15/1993 17:16:29   set iter=0 on process restore
//		J. Kercheval  Mon, 07/19/1993 10:07:57   exit and restart concurrent buffer if active
//		J. Kercheval  Fri, 07/23/1993 18:31:27   fix logic problem restarting concurrent buffer
//		J. Kercheval  Tue, 07/27/1993 16:56:40   set has_arg to FALSE
//		J. Kercheval  Wed, 09/08/1993 17:57:17   added a -y to the vmrg command line
//		J. Kercheval  Wed, 09/08/1993 19:51:03   save and restore the current process point and mark
//		J. Kercheval  Fri, 09/16/1994 10:07:49   add TLIB support (and make new VCS systems easy to add)
//		J. Kercheval  Sun, 10/09/1994 21:46:23   add RCS support
//		J. Kercheval  Tue, 11/01/1994 23:21:21   add replace_in_readonly_hook
//		J. Kercheval  Thu, 11/10/1994 19:44:56   Final for Epsilon V7.0 B2
//		J. Kercheval  Wed, 11/16/1994 21:57:42   Do not capture put to avoid NT capture bug in V7.0 B2
//		J. Kercheval  Sat, 12/24/1994 09:45:55   Add a little better branch handling
//		J. Kercheval  Sat, 12/31/1994 08:53:34   Restore line translation in VCSEnd
//		J. Kercheval  Mon, 02/13/1995 20:59:44   Final V7.0 modifications
//		J. Kercheval  Sat, 02/18/1995 10:30:04   Add VCS_result_popup
//		J. Kercheval  Wed, 04/19/1995 21:49:13   Correct window lookup logic for results
//		J. Kercheval  Thu, 05/11/1995 20:52:31   Add SourceSafe support
//		J. Kercheval  Sun, 01/28/1996 12:39:16   Store vcs variables in session file
//		J. Kercheval  Sun, 01/28/1996 12:59:25   Set the current project when applicable
//		J. Kercheval  Tue, 04/16/1996 23:30:19   Make a custom version of do_push (VCSDoPush)
//		J. Kercheval  Wed, 06/26/1996 10:29:09   Make use of replace_in_existing_hook()
//		J. Kercheval  Mon, 05/26/1997 16:57:54   Update for use with V9.0 of Epsilon
//
/*
Changes by Lasse Kinnunen, 3 December 2002:

Added check-in in of inital version (was missing, only done for RCS)

Added unlock of locked file (was missing, done for all environments)

Removed check for file read-only at check-in time.
I found this check to be irrelevant, you may e.g. have copied
a version-handled read-only file on top of a checked-out file.

Changed interface slightly to avoid prompts except where mandatory.
put() checks in and unlocks,
putl() checks in and keeps file locked,
get() checks out latest version without locking,
getl() checks out latest version and locks
unlk() unlocks a previously locked file.
 */


#include <eel.h>
#include <proc.h>

#define BOOLEAN int
#define TRUE 1
#define FALSE 0

#define _A_NORMAL	0x00				// Normal file - No read/write restrictions
#define _A_RDONLY	0x01				// Read only file
#define _A_HIDDEN	0x02				// Hidden file
#define _A_SYSTEM	0x04				// System file
#define _A_VOLID	0x08				// Volume ID file
#define _A_SUBDIR	0x10				// Subdirectory
#define _A_ARCH		0x20				// Archive file

#define PROCESS_BUFFER					"process"
#define PROCESS_RESTORE_BUFFER			"-VCSOldProcess-"
#define PROCESS_RESULT_BUFFER			"vcs-result"


char _VCSBuffer[128];					// current buffer name
char _VCSFilename[128];					// current buffer associated file name
char _VCSEditPath[128];					// the path of the current epsilon session
char _VCSTargetPath[128];				// the path of the current buffer
char *_VCSTempBuf;						// a temp buffer to maintain the current window
int _VCSPoint;							// current buffer point
int _VCSMark;							// current buffer mark
BOOLEAN _VCSTranslate;					// current buffer is translating

int _VCSProcessErrorSpot;				// error_spot in process buffer
int _VCSProcessPoint;					// point for process buffer
int _VCSProcessMark;					// mark for process buffer
BOOLEAN _VCSProcessHadErrors;			// _had_errors in process buffer
BOOLEAN _VCSProcessLastWasGrep;			// _last_was_grep in process buffer
BOOLEAN _VCSRestoreProcessBuffer;		// true if the process buffer is stored
BOOLEAN _VCSRestartConcurrent;			// true if process buffer was concurrent
char _had_errors;						// used in proc.e


//
// The following defines and variable determine the Version Control
// System to use.
//
#define MAX_COMMAND_LENGTH 40
#define MAX_VCS_SYSTEMS 4

#define useAutoDetermine -1
#define useRCS		0			// RCS
#define useTLIB		1			// Burton's TLIB
#define usePVCS		2			// PVCS
#define useSS		3			// Visual Source Safe


user int VCS_result_popup = 0;			// 1 show the results in more mode
user int VCS_default = useAutoDetermine;// -1 auto, otherwise the enumerated type

user char VCS_project_base[FNAMELEN];	// root directory of VCS project
user char VCS_project_name[FNAMELEN];	// name of project at VCS root

char VCS_last_current_project[FNAMELEN]; // The last project set to the current project

int useVCSType;							// VCS used by this module


//
// The following definitions are the commands used for access to the VCS
// log files.
//
typedef struct {
    char vcsEnvironmentID[MAX_COMMAND_LENGTH];
    char vcsSetProject[MAX_COMMAND_LENGTH];
    char vcsListing[MAX_COMMAND_LENGTH];
    char vcsIsLocked[MAX_COMMAND_LENGTH];
    char vcsIsLockedOutputTag[MAX_COMMAND_LENGTH];
    char vcsPut[MAX_COMMAND_LENGTH];
    char vcsPutLockOption[MAX_COMMAND_LENGTH];
    char vcsPutUnlockOption[MAX_COMMAND_LENGTH];
    char vcsGet[MAX_COMMAND_LENGTH];
    char vcsGetLockOption[MAX_COMMAND_LENGTH];
    char vcsGetUnlockOption[MAX_COMMAND_LENGTH];
    char vcsDiff[MAX_COMMAND_LENGTH];
    char vcsMerge[MAX_COMMAND_LENGTH];
    char vcsRevisionTag[MAX_COMMAND_LENGTH];
    char vcsCollisionFlag[MAX_COMMAND_LENGTH];
    char vcsLock[MAX_COMMAND_LENGTH];
    char vcsUnlock[MAX_COMMAND_LENGTH];
} VCSCommandList;

VCSCommandList vcsCommandList[MAX_VCS_SYSTEMS] = {
    { // RCS
	"LOGNAME",
	"",
	"rlog %s",
	"rlog -l%s %s",
	"-----",
	"ci -j %s %s",
	"-l",
	"-u",
	"co -q -f %s %s",
	"-l",
	"-u",
	"rcsdiff -b -c %s",
	"rcsmerge -r%s -r %s",
	"$Revision: ",
	"<<<<<<<",
	"rcs -l %s",
	"rcs -u %s"
    },
    { // TLIB
	"TLIBID",
	"",
	"tlib L %s",
	"tlib T %s",
	"This entry not used for TLIB",
	"tlib U%s %s",
	"K",
	"",
	"tlib E%s %s",
	"",
	"B",
	"diff is not supported for TLIB",
	"diff3 merges are not supported for TLIB",
	"diff3 merges are not supported for TLIB",
	"diff3 merges are not supported for TLIB",
	"TLIB ER %s",
	"TLIB UD %s"
    },
    { // PVCS
	"VCSID",
	"",
	"vlog %s",
	"vlog -BL%s %s",
	"vlog -BL",
	"put -N %s %s",
	"-l",
	"-u",
	"get -q %s %s",
	"-l",
	"",
	"vdiff -b -R %s",
	"vmrg -y -R%s -R %s",
	"$Revision: ",
	">>>>>>>>>>",
	"vcs -l %s",
	"vcs -u %s"
    },
    { // SourceSafe
	"SSUSER",
	"ss CP %s",
	"ss History %s",
	"ss Status -u -p %s",
	"This entry not used for SourceSafe",
	"ss Update %s %s",
	"-k",
	"",
	"%s %s",
	"ss Checkout -gl.",
	"ss Get -w- -gl.",
	"ss Difference %s",
	"Merge is not supported for SourceSafe",
	"$Revision: ",
	"======",
	"ss Checkout -g- %s",
	"ss Uncheckout -g- %s"
    }
};


////////////////////////////////////////////////////////////////////////////
//
// IsFile() returns FALSE if the buffer does not have an associated
// filename
//
//
BOOLEAN IsFile(f)
char *f;
{
    return strlen(f);
}


////////////////////////////////////////////////////////////////////////////
//
// IsWritable() returns FALSE if the file is write only and TRUE
// otherwise
//
//
BOOLEAN IsWritable(f)
char *f;
{
    struct file_info fi;

    //
    // validate a valid parameter
    //
    if (!IsFile(f)) {
	return FALSE;
    }

    //
    // check the file attribute, if it doesn't exist then it is writable
    //
    if (!check_file(f, &fi)) {
	return TRUE;
    }

    if (fi.attr & _A_RDONLY &&
	  !(fi.attr & _A_SUBDIR)) {
	return FALSE;
    } else {
	return TRUE;
    }
}


////////////////////////////////////////////////////////////////////////////
//
// do_branch() removes the readonly flag from the current buffer's file
// after updating the file.
//
//
void do_branch(checkLog)
BOOLEAN checkLog;
{
    char prompt[80];
    char response[80];
    char filename_relative[80];
    char *str;

    int result = 0;

    save_var shell_shrinks = TRUE;

    //
    // Determine the VCS system and set it to the current project
    //
    if (!VCSSetCurrentProject()) {
	return;
    }

    //
    // obtain the relative filename
    //
    relative(filename, filename_relative);

    //
    // obtain the current edit path and change to target path
    //
    getcd(_VCSEditPath);
    str = get_tail(filename, TRUE);
    strncpy(_VCSTargetPath, filename, strlen(filename)-strlen(str));

    //
    // save the current process buffer
    //
    chdir(_VCSTargetPath);
    VCSProcessSave();

    //
    // check if the file is locked
    //
    if (checkLog &&
	  IsLocked(filename)) {

	//
	// restore for the query
	//
	chdir(_VCSEditPath);
	VCSProcessRestore();

	//
	// if locked ask if want unlocked
	//
	sprintf(prompt, "%s is already locked, continue?", filename_relative);
	get_strdef(response, prompt, "n");
	if (!strnfcmp(response, "y", 1)) {
	    //
	    // perform the unlock
	    //
	    VCSProcessSave();
	    sprintf(prompt, vcsCommandList[useVCSType].vcsUnlock, get_tail(filename_relative, TRUE));
	    shell_shrinks = TRUE;
	    result = VCSDoPush(prompt, TRUE, FALSE);
	    VCSProcessRestore();
	} else {
	    //
	    // do not continue
	    //
	    return;
	}

	//
	// reset for the branch
	//
	chdir(_VCSTargetPath);
	VCSProcessSave();
    }

    if ((!result || !checkLog) &&
	  !IsWritable(filename)) {
	//
	// chmod the file to be writeable
	//
	sprintf(prompt, "attrib -r %s", get_tail(filename_relative, TRUE));
	result = VCSDoPush(prompt, TRUE, FALSE);
    }

    chdir(_VCSEditPath);
    VCSProcessRestore();

    //
    // view the results
    //
    if (result) {
	showResults(PROCESS_RESULT_BUFFER);
    } else {
	set_read_only(FALSE);
	if (checkLog) {
	    note("%s branched.", filename_relative);
	}
    }
}


////////////////////////////////////////////////////////////////////////////
//
// do_get() will perform an update from PVCS on the current buffer's
// file and will perform a lock or not depending upon the lock
// parameter
//
//
void do_get(lock)
BOOLEAN lock;
{
    char response[80];
    char prompt[80];
    char filename_relative[80];
    char options[80];
    int result = 0;

    //
    // Determine the VCS system and set it to the current project
    //
    if (!VCSSetCurrentProject()) {
	return;
    }

    //
    // obtain the relative filename
    //
    relative(filename, filename_relative);

    //
    // if the current file is writable then verify the overwrite,
    // else exit
    //
    if (useVCSType != useTLIB &&
	  IsWritable(filename)) {
	sprintf(prompt, "%s is writeable, continue?", filename_relative);
	get_strdef(response, prompt, "y");
	if (strnfcmp(response, "y", 1)) {
	    return;
	}
	if (useVCSType == useSS) {
	    VCSProcessSave();
	    sprintf(prompt, "attrib +r %s", get_tail(filename_relative, TRUE));
	    result = VCSDoPush(prompt, TRUE, FALSE);
	    VCSProcessRestore();
	}
    }

    //
    // determine the lock option
    //
    if (lock) {
	strcpy(options, vcsCommandList[useVCSType].vcsGetLockOption);
    } else {
	strcpy(options, vcsCommandList[useVCSType].vcsGetUnlockOption);
    }

    if (VCSStart()) {

	//
	// perform the get
	//
	sprintf(prompt, vcsCommandList[useVCSType].vcsGet, options, get_tail(filename_relative, TRUE));
	shell_shrinks = TRUE;
	result += VCSDoPush(prompt, TRUE, FALSE);

	VCSEnd();

	//
	// view the results
	//
	if (result) {
	    showResults(PROCESS_RESULT_BUFFER);
	} else {
	    if (lock) {
		note("LOCK of %s complete.", filename_relative);
	    } else {
		note("GET of %s complete.", filename_relative);
	    }
	}
    }
}


////////////////////////////////////////////////////////////////////////////
//
// do_merge() will merge the current buffer's file with the base
// version found using PVCS.
//
//
do_merge(lock)
BOOLEAN lock;
{
    char prompt[80];
    char response[80];
    char revision[80];
    char filename_relative[80];
    char backup_filename[80];
    char original_filename[80];
    int result = 0;
    int previousPoint;

    save_var shell_shrinks = TRUE;

    //
    // Determine the VCS system and set it to the current project
    //
    if (!VCSSetCurrentProject()) {
	return;
    }

    //
    // merge is not supported for TLIB
    //
    if (useVCSType == useTLIB) {
	note("Not supported when using TLIB.");
	return;
    }

    //
    // merge is not supported for SourceSafe
    //
    if (useVCSType == useSS) {
	note("Not supported when using SourceSafe.");
	return;
    }

    //
    // obtain the relative filename
    //
    relative(filename, filename_relative);

    //
    // if the current file is not writable then scold the user
    //
    if (!IsWritable(filename)) {
	note("%s must be a writeable file.", filename_relative);
	return;
    }

    //
    // should we lock the file before merging?
    //
    if (!lock) {
	get_strdef(response, "Lock the merged file?", "n");
	if (!strnfcmp(response, "y", 1)) {
	    lock = TRUE;
	}
    }

    //
    // obtain the file revision
    //
    if (!VCSGetRevision(revision)) {
	say("%s has no revision field, merge aborted.", filename);
	return;
    }

    //
    // perform the merge
    //
    if (VCSStart()) {
	//
	// copy the file to a backup location
	//
	build_filename(backup_filename, backup_name,
		       get_tail(filename_relative, TRUE));
	strcpy(original_filename, filename);
	filename = backup_filename;
	do_save_file(1, 0, 0);
	filename = original_filename;

	//
	// lock the file if needed
	//
	if (lock) {
	    sprintf(prompt, vcsCommandList[useVCSType].vcsLock, get_tail(filename_relative, TRUE));
	    shell_shrinks = TRUE;
	    result = VCSDoPush(prompt, TRUE, FALSE);

	    //
	    // if the lock fails then query
	    //
	    if (result) {
		//
		// restore for the query
		//
		lock = FALSE;
		VCSEnd();

		//
		// query
		//
		get_strdef(response, "File could not be locked, merge anyway?", "y");
		if (strnfcmp(response, "y", 1)) {
		    return;
		}

		//
		// restart
		//
		VCSStart();
	    }
	}

	//
	// perform the merge
	//
	sprintf(prompt, vcsCommandList[useVCSType].vcsMerge, revision, get_tail(filename_relative, TRUE));
	result = VCSDoPush(prompt, TRUE, FALSE);

	//
	// RCS fails with return value 2 or greater, otherwise
	// failure is non-zero return value.
	//
	if (((useVCSType == useRCS && result > 1) ||
	     (useVCSType == usePVCS && result)) &&
	      lock) {
	    //
	    // unlock if the merge failed
	    //
	    sprintf(prompt, vcsCommandList[useVCSType].vcsUnlock, get_tail(filename_relative, TRUE));
	    shell_shrinks = TRUE;
	    VCSDoPush(prompt, TRUE, FALSE);
	}

	VCSEnd();
    } else {
	return;
    }

    if ((useVCSType == useRCS && result > 1) ||
	  (useVCSType == usePVCS && result)) {
	showResults(PROCESS_RESULT_BUFFER);
    } else {
	//
	// grep the file for conflicts
	//
	previousPoint = point;
	point = 0;
	if (search(1, vcsCommandList[useVCSType].vcsCollisionFlag)) {
	    _last_was_grep = TRUE;
	    do_grep("grep", vcsCommandList[useVCSType].vcsCollisionFlag, 0, filename, 0);
	    next_position();
	} else {
	    point = previousPoint;
	}
	if (lock) {
	    say("%s merged and locked.", filename);
	} else {
	    say("%s merged.", filename);
	}
    }
}


////////////////////////////////////////////////////////////////////////////
//
// do_put() will take the current buffer's file, check it in and unlock
// the logfile.  This assumes the logfile is currently locked and
// merged.
//
//
void do_put(lock)
BOOLEAN lock;
{
    char prompt[80];
    char response[80];
    char filename_relative[80];
    char options[80];
    int result;

    save_var shell_shrinks = TRUE;

    //
    // Determine the VCS system and set it to the current project
    //
    if (!VCSSetCurrentProject()) {
	return;
    }

    //
    // obtain the relative filename
    //
    relative(filename, filename_relative);

    //
    // should we lock the file once it is put?
    //
    if (lock == TRUE) {
	strcpy(options, vcsCommandList[useVCSType].vcsPutLockOption);
    } else {
	strcpy(options, vcsCommandList[useVCSType].vcsPutUnlockOption);
    }

    //
    // perform the put
    //
    if (VCSStart()) {
	sprintf(prompt, vcsCommandList[useVCSType].vcsPut, options, get_tail(filename_relative, TRUE));
	result = VCSDoPush(prompt, FALSE, -1);

	VCSEnd();

	//
	// display the results
	//
	if (result) {
	    if (useVCSType == useRCS) {
		sprintf( prompt, "Create initial version of %s ?", filename_relative);
		get_strdef(response, prompt, "y");

		if (strnfcmp(response, "y", 1) == 0) {
		    if (lock) {
			sprintf( prompt, "ci -i -l %s", filename_relative);
		    } else {
			sprintf( prompt, "ci -i -u %s", filename_relative);
		    }

		    result = VCSDoPush(prompt, FALSE, -1);

		    if (result) {
			note("Error creating initial version of %s", filename_relative);
		    } else if (lock) {
			note("PUT/Create and lock of %s complete.", filename_relative);
		    } else {
			note("PUT/Create of %s complete.", filename_relative);
		    }
		}
	    } else {
		note("Error during PUT of %s", filename_relative);
	    }
	} else if (lock) {
	    note("PUT and lock of %s complete.", filename_relative);
	} else {
	    note("PUT of %s complete.", filename_relative);
	}
    }
}


////////////////////////////////////////////////////////////////////////////
//
// showResults will display the vcs result buffer appropriately
//
//
void showResults(buf)
char *buf;
{
    save_var bufname;

    if (VCS_result_popup) {
	view_buffer(buf, FALSE);
    } else {
	quiet_set_bookmark();
	bufname = buf;
	point = 0;
	locate_window(buf, "");
    }
}


////////////////////////////////////////////////////////////////////////////
//
// do_unlock() unlocks the current buffer's file.
//
//
void do_unlock()
{
    char prompt[80];
    char response[80];
    char filename_relative[80];
    char *str;

    int result = 0;

    save_var shell_shrinks = TRUE;

    //
    // Determine the VCS system and set it to the current project
    //
    if (!VCSSetCurrentProject()) {
	return;
    }

    //
    // obtain the relative filename
    //
    relative(filename, filename_relative);

    //
    // obtain the current edit path and change to target path
    //
    getcd(_VCSEditPath);
    str = get_tail(filename, TRUE);
    strncpy(_VCSTargetPath, filename, strlen(filename)-strlen(str));

    //
    // save the current process buffer
    //
    chdir(_VCSTargetPath);
    VCSProcessSave();

    //
    // check if the file is locked
    //
    if (IsLocked(filename)) {

	//
	// restore for the query
	//
	chdir(_VCSEditPath);
	VCSProcessRestore();

	//
	// if locked ask if want unlocked
	//
	sprintf(prompt, "Unlock %s ?", filename_relative);
	get_strdef(response, prompt, "y");
	if (strnfcmp(response, "y", 1) == 0) {
	    //
	    // perform the unlock
	    //
	    VCSProcessSave();
	    sprintf(prompt, vcsCommandList[useVCSType].vcsUnlock, get_tail(filename_relative, TRUE));
	    shell_shrinks = TRUE;
	    result = VCSDoPush(prompt, TRUE, FALSE);
	    VCSProcessRestore();

	    if (result) {
		showResults(PROCESS_RESULT_BUFFER);
	    } else {
		note("%s unlocked.", filename_relative);
	    }
	} else {
	    return;
	}
    }
}


////////////////////////////////////////////////////////////////////////////
//
// VCSDetermineVCS() will determine the VCS in use.  If the default
// is -1 then the VCS is determined by looking for the ID environment
// variable for each VCS available in order.  If no variable is found
// during this phase then the first VCS in the command list is used.
//
VCSDetermineVCS()
{
    int index;
    char *vcsid;

    useVCSType = VCS_default;
    if (useVCSType >= MAX_VCS_SYSTEMS ||
	  useVCSType <= useAutoDetermine) {
	useVCSType = 0;
    }
    if (VCS_default == useAutoDetermine) {
	for (index = 0; index < MAX_VCS_SYSTEMS; index++) {
	    vcsid = getenv(vcsCommandList[index].vcsEnvironmentID);
	    if (vcsid) {
		useVCSType = index;
		break;
	    }
	}
    }
}


////////////////////////////////////////////////////////////////////////////
//
// VCSProcessSave() save any current process buffer.  VERY few
// changes to do_push().  Just copy it over on updates and reedit
// changes in.
//
//
VCSProcessSave()
{
    save_var bufname;

    //
    // save the process buffer if there is one currently around
    //
    _VCSRestoreProcessBuffer = FALSE;
    if (exist(PROCESS_BUFFER)) {
	_VCSRestoreProcessBuffer = TRUE;
	_VCSProcessHadErrors = _had_errors;
	_VCSProcessLastWasGrep = _last_was_grep;
	_VCSProcessPoint = point;
	_VCSProcessMark = mark;
	if (spot_to_buffer(error_spot) > 0) {
	    _VCSProcessErrorSpot = *error_spot;
	} else {
	    _VCSProcessErrorSpot = 0;
	}
	_VCSRestartConcurrent =	another_process() ? TRUE : FALSE;
	try_exit_concurrent();
	delete_buffer(PROCESS_RESTORE_BUFFER);
	bufname = PROCESS_BUFFER;
	change_buffer_name(PROCESS_RESTORE_BUFFER);
    }
}


////////////////////////////////////////////////////////////////////////////
//
// VCSProcessRestore() restore the current process buffer
//
//
VCSProcessRestore()
{
    save_var bufname;

    //
    // move the process buffer to the results buffer
    //
    if (exist(PROCESS_BUFFER)) {
	delete_buffer(PROCESS_RESULT_BUFFER);
	bufname = PROCESS_BUFFER;
	change_buffer_name(PROCESS_RESULT_BUFFER);
    }

    //
    // restore the previous process buffer
    //
    if (_VCSRestoreProcessBuffer) {
	delete_buffer(PROCESS_BUFFER);
	bufname = PROCESS_RESTORE_BUFFER;
	change_buffer_name(PROCESS_BUFFER);
	if (_VCSRestartConcurrent) {
	    maybe_restart_concurrent();
	}
	_had_errors = _VCSProcessHadErrors;
	_last_was_grep = _VCSProcessLastWasGrep;

	//
	// re-establish the error spot
	//
	point =	_VCSProcessErrorSpot <= size() ? _VCSProcessErrorSpot : size();
	set_error_spot();

	//
	// re-establish the point and the mark
	//
	point =	_VCSProcessPoint <= size() ? _VCSProcessPoint : size();
	mark = _VCSProcessMark <= size() ? _VCSProcessMark : size();
    }

    iter = 0;
    has_arg = 0;
}


////////////////////////////////////////////////////////////////////////////
//
// VCSDoPush() performs a simpler and kinder do_push.  This push does
// not modify the current directory for Epsilon at all (the variable
// start_process_in_buffer_directory is completely ignored).  This
// function will also not abort on error.
//
//
int VCSDoPush(cmdline, cap, show)
char *cmdline;	// cap nonzero means capture output, show means show it to user
{		// but show -1 means don't wait for keystroke
    char *s, dir[FNAMELEN], old[FNAMELEN];
    int result, tbuf;

    if (!another_process())
	save_var restart_concurrent = 0;
    if (opsys == OS_DOS && !is_gui)
	try_exit_concurrent();
    iter = 1;
    before_push();
    getcd(old);
    get_buffer_directory(dir);
    chdir(dir);
    note(cmdline);
    result = shell("", cmdline, cap ? bufnum_to_name(tbuf = tmp_buf()) : "");
    chdir(old);
    if (cap) {
	process_captured(tbuf, cmdline);
	buf_delete(tbuf);
    } else if (cmdline[0] && show != -1 && !is_gui) {
	if (!char_avail()) {
	    s = "Press any key to return to Epsilon";
	    term_write(0, screen_lines - 1, s, strlen(s),
		       color_class after_exiting, 1);
	    term_position(strlen(s), screen_lines - 1);
	}
	getkey();
    }
    after_push();
    if (cmdline[0] && cap && show) {
	locate_window(PROCBUF, "");
	point = size();
    }
    maybe_restart_concurrent();
    if (cmdline[0] && result)
	say("Process returned %d", result);
    else
	check_dates(0);
    return result;
}


////////////////////////////////////////////////////////////////////////////
//
// VCSSetCurrentProject() sets the project based on the base and
// relative path from the base.
//
//
BOOLEAN VCSSetCurrentProject()
{
    //
    // validate this buffer represents a file
    //
    if (!IsFile(filename)) {
	note("This buffer does not represent a file.");
	return FALSE;
    }

    //
    // determine which VCS system we are using
    //
    VCSDetermineVCS();

    //
    // Do no work unless necessary
    //
    if (strlen(vcsCommandList[useVCSType].vcsSetProject) &&
	  strlen(VCS_project_name)) {

	char currentProject[FNAMELEN];
	char tmpStr[FNAMELEN];
	char tmpFile[FNAMELEN];
	BOOLEAN result;

	//
	// Compute the new current project path
	//
	if (strlen(VCS_project_base)) {
	    strcpy(tmpStr, VCS_project_base);
	} else {
	    give_session_file(tmpFile);
	    build_filename(tmpStr, "%p", tmpFile);
	}
	absolute(tmpStr);
	if (strncmp(tmpStr, filename, strlen(tmpStr))) {
	    note("File does not reside in VCS project subdirectory tree.");
	    return FALSE;
	}
	strcpy(currentProject, VCS_project_name);
	strcpy(tmpFile, filename);
	tmpFile[strlen(filename) - strlen(get_tail(filename, TRUE)) - 1] = '\0';
	strcat(currentProject, &tmpFile[strlen(tmpStr)]);

	//
	// if the new path is the same as the previous one then no
	// need to do it again.
	//
	if (strcmp(currentProject, VCS_last_current_project)) {
	    //
	    // If the set fails then warn the user and return false
	    //
	    VCSProcessSave();
	    sprintf(tmpStr, vcsCommandList[useVCSType].vcsSetProject, currentProject);
	    result = VCSDoPush(tmpStr, TRUE, FALSE);
	    VCSProcessRestore();
	    if (result) {
		note("Unable to set the current VCS project to %s.", currentProject);
		return FALSE;
	    } else {
		strcpy(VCS_last_current_project, currentProject);
	    }
	}
    }

    return TRUE;
}


////////////////////////////////////////////////////////////////////////////
//
// VCSStart() validates a buffer and saves the current buffer state
//
//
BOOLEAN VCSStart()
{
    char *str;

    //
    // save the current buffer state and file
    //
    _VCSPoint = point;
    _VCSMark = mark;
    _VCSTranslate = translation_type;
    strcpy(_VCSBuffer, bufname);
    strcpy(_VCSFilename, filename);
    if (modified && IsWritable(filename)) {
	do_save_file(0, 1, 1);
    }

    //
    // obtain the current edit path and change to target path
    //
    getcd(_VCSEditPath);
    str = get_tail(filename, TRUE);
    strcpy(_VCSTargetPath, filename);
    _VCSTargetPath[strlen(filename) - strlen(str)] = '\0';
    chdir(_VCSTargetPath);

    //
    // create a temporary buffer and use it to maintain the window
    //
    _VCSTempBuf = temp_buf();
    to_buffer(_VCSTempBuf);

    //
    // remove the current file buffer
    //
    delete_buffer(_VCSBuffer);

    //
    // tuck away the old existing process buffer if it exists
    //
    VCSProcessSave();

    return TRUE;
}


////////////////////////////////////////////////////////////////////////////
//
// VCSEnd() restores a buffer
//
//
BOOLEAN VCSEnd()
{
    //
    // restore the target buffers vital stats
    //
    find_in_other_buf(_VCSFilename, _VCSTranslate);
    point = _VCSPoint <= size() ? _VCSPoint : size();
    mark = _VCSMark <= size() ? _VCSMark : size();
    chdir(_VCSEditPath);
    delete_buffer(_VCSTempBuf);

    //
    // save the current process and restore the previous one
    //
    VCSProcessRestore();

    return TRUE;
}


////////////////////////////////////////////////////////////////////////////
//
// IsLocked() returns TRUE if the filename represents a locked file.
//
//
BOOLEAN IsLocked(f)
char *f;
{
    char prompt[80];
    char *vcsid;
    BOOLEAN result;
    BOOLEAN rval = FALSE;
    int previousPoint;

    save_var bufname;

    if (useVCSType == useTLIB ||
	  useVCSType == useSS) {
	sprintf(prompt, vcsCommandList[useVCSType].vcsIsLocked, get_tail(f, TRUE));
	shell_shrinks = TRUE;
	result = VCSDoPush(prompt, TRUE, FALSE);

	rval = result ? TRUE : FALSE;
	if (useVCSType == useTLIB) {
	    rval = !rval;
	}
    } else {
	vcsid = getenv(vcsCommandList[useVCSType].vcsEnvironmentID);
	if (vcsid) {
	    //
	    // check for a lock
	    //
	    sprintf(prompt, vcsCommandList[useVCSType].vcsIsLocked, vcsid, get_tail(f, TRUE));
	    shell_shrinks = TRUE;
	    result = VCSDoPush(prompt, TRUE, FALSE);

	    //
	    // examine the results
	    //
	    if (!result) {
		bufname = PROCESS_BUFFER;
		previousPoint = point;
		search(-1, vcsCommandList[useVCSType].vcsIsLockedOutputTag);
		point = matchstart;
		nl_forward();
		if (search(1, vcsid)) {
		    rval = TRUE;
		}
		point = previousPoint;
	    }
	}
    }

    return rval;
}


////////////////////////////////////////////////////////////////////////////
//
// VCSGetRevision() obtains the revision number of the current file.
// This routine will return FALSE on error.
//
//
BOOLEAN VCSGetRevision(revision)
char *revision;
{
    BOOLEAN rval = FALSE;

    save_spot point;

    point = 0;
    revision[0] = '\0';
    if (search(1, vcsCommandList[useVCSType].vcsRevisionTag)) {
	point = matchend;
	for (;curchar() == ' ' && point < size(); point++);
	grab_string(bufnum, revision, " ");
	if (strlen(revision)) {
	    rval = TRUE;
	}
    }

    return rval;
}


////////////////////////////////////////////////////////////////////////////
//
// make_session_hook and restore_session_hook save the current
// project variables to the session file when it is written or read
// and then call any other defined session hooks.
//

#define VCS_START_DELIM	"<<SES!@!VCS_START>>"
#define VCS_END_DELIM	"<<SES!@!VCS_END>>"

new_vcs_make_session_hook(b)
{
    save_var bufnum = b;

    point = size();

    buf_printf(b, "\n%s", VCS_START_DELIM);
    buf_printf(b, "\n%d", VCS_result_popup);
    buf_printf(b, "\n%d", VCS_default);
    buf_printf(b, "\n%s", VCS_project_base);
    buf_printf(b, "\n%s", VCS_project_name);
    buf_printf(b, "\n%s\n", VCS_END_DELIM);

    vcs_make_session_hook(b);
}

REPLACE_FUNC("vcs", "make_session_hook")

new_vcs_restore_session_hook(b)
{
    int old_point = point;

    save_var bufnum = b;

    //
    // Init variables
    //
    VCS_last_current_project[0] = '\0';

    //
    // obtain the project variables
    //
    point = 0;
    if (!search(1, VCS_START_DELIM)) {
	point = old_point;
    } else {
	nl_forward();			 // move to first variable
	grab_numbers(b, &VCS_result_popup);
	grab_numbers(b, &VCS_default);
	grab_line(b, VCS_project_base);
	grab_line(b, VCS_project_name);
	nl_forward();			 // move beyond final delimiter
    }

    //
    // call the previous hook
    //
    vcs_restore_session_hook(b);
}

REPLACE_FUNC("vcs", "restore_session_hook")


////////////////////////////////////////////////////////////////////////////
//
// get() will perform an update from PVCS on the current buffer's file
//
//
command get()
{
    do_get(FALSE);
}


////////////////////////////////////////////////////////////////////////////
//
// lock() will update the current buffer's file and lock the file
// using PVCS
//
//
command getl()
{
    do_get(TRUE);
}

////////////////////////////////////////////////////////////////////////////
//
// branch() preforms a branch and checks the current log file to
// remove any current locks.
//
//
command branch()
{
    do_branch(TRUE);
}



////////////////////////////////////////////////////////////////////////////
//
// merge() calls do_merge with a default not to lock the file
//
//
command merge()
{
    do_merge(FALSE);
}


////////////////////////////////////////////////////////////////////////////
//
// mergel() calls do_merge with the parameter to lock the file
//
//
command mergel()
{
    do_merge(TRUE);
}


////////////////////////////////////////////////////////////////////////////
//
// put() will take the current buffer's file, check it in and unlock
// the logfile.  This assumes the logfile is currently locked and
// merged.
//
//
command put()
{
    do_put( FALSE);
}

////////////////////////////////////////////////////////////////////////////
//
// putl() will take the current buffer's file, check it in and keep
// the logfile locked.  This assumes the logfile is currently locked and
// merged.
//
//
command putl()
{
    do_put( TRUE);
}

////////////////////////////////////////////////////////////////////////////
//
// unlk() will unlock the current buffer's file.
//
//
command unlk()
{
    do_unlock();
}

////////////////////////////////////////////////////////////////////////////
//
// vlog() will show the log for the current buffer's file.
//
//
command vlog()
{
    char prompt[80];
    char filename_relative[80];

    save_var shell_shrinks = TRUE;

    //
    // Determine the VCS system and set it to the current project
    //
    if (!VCSSetCurrentProject()) {
	return;
    }

    //
    // obtain the relative filename
    //
    relative(filename, filename_relative);

    VCSStart();

    //
    // run the shell
    //
    sprintf(prompt, vcsCommandList[useVCSType].vcsListing, get_tail(filename_relative, TRUE));
    VCSDoPush(prompt, TRUE, FALSE);

    VCSEnd();

    //
    // show the process buffer in more mode
    //
    showResults(PROCESS_RESULT_BUFFER);
}


////////////////////////////////////////////////////////////////////////////
//
// vdiff() will show the difference between the current buffer's file
// and the current logfile contents.
//
//
command vdiff()
{
    char prompt[80];
    char filename_relative[80];

    save_var shell_shrinks = TRUE;

    //
    // Determine the VCS system and set it to the current project
    //
    if (!VCSSetCurrentProject()) {
	return;
    }

    //
    // vdiff is not supported for TLIB
    //
    if (useVCSType == useTLIB) {
	note("Not supported when using TLIB.");
	return;
    }

    //
    // obtain the relative filename
    //
    relative(filename, filename_relative);

    VCSStart();

    //
    // run the shell
    //
    sprintf(prompt, vcsCommandList[useVCSType].vcsDiff, get_tail(filename_relative, TRUE));
    VCSDoPush(prompt, TRUE, FALSE);

    VCSEnd();

    //
    // show the process buffer in more mode
    //
    showResults(PROCESS_RESULT_BUFFER);
}


////////////////////////////////////////////////////////////////////////////
//
// replace_in_readonly_hook() and replace_in_existing_hook() will
// branch the file if it is readonly.  This is the simplest error
// handling and does not require user intervention if the file can
// not be locked or the log files are not currently available.  This
// is also the most general solution.
//
//
new_vcs_replace_in_readonly_hook(old_readonly)
{
    //
    // silence the compiler warning
    //
    old_readonly = old_readonly;

    //
    // branch the file for use
    //
    do_branch(FALSE);
}

REPLACE_FUNC("vcs", "replace_in_readonly_hook")

new_vcs_replace_in_existing_hook(old_readonly)
{
    //
    // silence the compiler warning
    //
    old_readonly = old_readonly;

    //
    // branch the file for use
    //
    do_branch(FALSE);
}

REPLACE_FUNC("vcs", "replace_in_existing_hook")
